@@ -30,17 +30,20 @@
 #include <gmodule.h>
 #include "girepository.h"
 #include "gtypelib.h"
+#include "ginfo.h"
 
 #include "config.h"
 
 static GStaticMutex globals_lock = G_STATIC_MUTEX_INIT;
 static GIRepository *default_repository = NULL;
 static GSList *search_path = NULL;
+static GSList *override_search_path = NULL;
 
 struct _GIRepositoryPrivate 
 {
   GHashTable *typelibs; /* (string) namespace -> GTypelib */
   GHashTable *lazy_typelibs; /* (string) namespace-version -> GTypelib */
+  GHashTable *info_by_gtype; /* GType -> GIBaseInfo */
 };
 
 G_DEFINE_TYPE (GIRepository, g_irepository, G_TYPE_OBJECT);
@@ -56,6 +59,10 @@ g_irepository_init (GIRepository *repository)
 			     (GDestroyNotify) g_typelib_free);
   repository->priv->lazy_typelibs 
     = g_hash_table_new (g_str_hash, g_str_equal);
+  repository->priv->info_by_gtype
+    = g_hash_table_new_full (g_direct_hash, g_direct_equal,
+                             (GDestroyNotify) NULL,
+                             (GDestroyNotify) g_base_info_unref);
 }
 
 static void
@@ -65,6 +72,7 @@ g_irepository_finalize (GObject *object)
 
   g_hash_table_destroy (repository->priv->typelibs);
   g_hash_table_destroy (repository->priv->lazy_typelibs);
+  g_hash_table_destroy (repository->priv->info_by_gtype);
   
   (* G_OBJECT_CLASS (g_irepository_parent_class)->finalize) (G_OBJECT (repository));
 }
@@ -82,7 +90,7 @@ g_irepository_class_init (GIRepositoryClass *class)
 }
 
 static void
-init_globals ()
+init_globals (void)
 {
   g_static_mutex_lock (&globals_lock);
 
@@ -97,9 +105,13 @@ init_globals ()
       char *typelib_dir;
       const gchar *type_lib_path_env;
 
+      /* This variable is intended to take precedence over both the default
+       * search path, as well as anything written into code with g_irepository_prepend_search_path.
+       */
       type_lib_path_env = g_getenv ("GI_TYPELIB_PATH");
 
       search_path = NULL;
+      override_search_path = NULL;
       if (type_lib_path_env)
         {
           gchar **custom_dirs;
@@ -110,7 +122,7 @@ init_globals ()
           d = custom_dirs;
           while (*d)
             {
-              search_path = g_slist_prepend (search_path, *d);
+              override_search_path = g_slist_prepend (override_search_path, *d);
               d++;
             }
 
@@ -120,7 +132,7 @@ init_globals ()
 
       libdir = GOBJECT_INTROSPECTION_LIBDIR;
 
-      typelib_dir = g_build_filename (libdir, "girepository", NULL);
+      typelib_dir = g_build_filename (libdir, "girepository-1.0", NULL);
 
       search_path = g_slist_prepend (search_path, typelib_dir);
 
@@ -137,6 +149,36 @@ g_irepository_prepend_search_path (const char *directory)
   search_path = g_slist_prepend (search_path, g_strdup (directory));
 }
 
+/**
+ * g_irepository_get_search_path:
+ *
+ * Returns the search path the GIRepository will use when looking for typelibs.
+ * The string is internal to GIRespository and should not be freed, nor should
+ * the elements.
+ *
+ * Return value: (element-type filename) (transfer none): list of strings
+ */
+GSList *
+g_irepository_get_search_path (void)
+{
+  return search_path;
+}
+
+static
+GSList *
+build_search_path_with_overrides (void)
+{
+	GSList *result;
+	if (override_search_path != NULL) 
+	  {
+		result = g_slist_copy (override_search_path);
+		g_slist_last (result)->next = g_slist_copy (search_path);
+	  } 
+	else
+	  result = g_slist_copy (search_path);
+	return result;
+}
+
 static char *
 build_typelib_key (const char *name, const char *source)
 {
@@ -164,13 +206,12 @@ get_typelib_dependencies (GTypelib *typelib)
 static GIRepository *
 get_repository (GIRepository *repository)
 {
+  init_globals ();
+
   if (repository != NULL)
     return repository;
-  else 
-    {
-      init_globals ();
-      return default_repository;
-    }
+  else
+    return default_repository;
 }
 
 static GTypelib *
@@ -432,7 +473,7 @@ g_irepository_is_registered (GIRepository *repository,
  * parameter to mean this default repository, which is usually more
  * convenient for C.
  * 
- * Returns: The global singleton #GIRepository 
+ * Returns: (transfer none): The global singleton #GIRepository 
  */
 GIRepository * 
 g_irepository_get_default (void)
@@ -473,6 +514,7 @@ g_irepository_get_n_infos (GIRepository *repository,
 
 typedef struct
 {
+  GIRepository *repo;
   gint index;
   const gchar *name;
   const gchar *type;
@@ -542,8 +584,9 @@ find_interface (gpointer key,
   if (index != 0)
     {
       entry = g_typelib_get_dir_entry (typelib, index);
-      iface_data->iface = g_info_new (entry->blob_type, NULL,
-				      typelib, entry->offset);
+      iface_data->iface = g_info_new_full (entry->blob_type,
+					   iface_data->repo,
+					   NULL, typelib, entry->offset);
     }
 }
 
@@ -571,6 +614,7 @@ g_irepository_get_info (GIRepository *repository,
 
   repository = get_repository (repository);
 
+  data.repo = repository;
   data.name = NULL;
   data.type = NULL;
   data.index = index + 1;
@@ -607,6 +651,13 @@ g_irepository_find_by_gtype (GIRepository *repository,
 
   repository = get_repository (repository);
 
+  data.iface = g_hash_table_lookup (repository->priv->info_by_gtype,
+                                    (gpointer)type);
+
+  if (data.iface)
+    return g_base_info_ref (data.iface);
+
+  data.repo = repository;
   data.name = NULL;
   data.type = g_type_name (type);
   data.index = -1;
@@ -615,6 +666,12 @@ g_irepository_find_by_gtype (GIRepository *repository,
   g_hash_table_foreach (repository->priv->typelibs, find_interface, &data);
   g_hash_table_foreach (repository->priv->lazy_typelibs, find_interface, &data);
 
+  if (data.iface)
+    g_hash_table_insert (repository->priv->info_by_gtype,
+                         (gpointer) type,
+                         g_base_info_ref (data.iface));
+
+
   return data.iface;
 }
 
@@ -643,6 +700,7 @@ g_irepository_find_by_name (GIRepository *repository,
 
   repository = get_repository (repository);
 
+  data.repo = repository;
   data.name = name;
   data.type = NULL;
   data.index = -1;
@@ -805,6 +863,7 @@ find_namespace_version (const gchar  *namespace,
 			const gchar  *version,
 			gchar       **path_ret)
 {
+  GSList *tmp_path;
   GSList *ldir;
   GError *error = NULL;
   GMappedFile *mfile = NULL;
@@ -812,7 +871,8 @@ find_namespace_version (const gchar  *namespace,
  
   fname = g_strdup_printf ("%s-%s.typelib", namespace, version);
 
-  for (ldir = search_path; ldir; ldir = ldir->next)
+  tmp_path = build_search_path_with_overrides ();
+  for (ldir = tmp_path; ldir; ldir = ldir->next)
     {
       char *path = g_build_filename (ldir->data, fname, NULL);
       
@@ -827,6 +887,7 @@ find_namespace_version (const gchar  *namespace,
       break;
     }
   g_free (fname);
+  g_slist_free (tmp_path);
   return mfile;
 }
 
@@ -925,6 +986,7 @@ find_namespace_latest (const gchar  *namespace,
 		       gchar       **version_ret,
 		       gchar       **path_ret)
 {
+  GSList *tmp_path;
   GSList *ldir;
   GError *error = NULL;
   char *namespace_dash;
@@ -940,7 +1002,8 @@ find_namespace_latest (const gchar  *namespace,
   namespace_typelib = g_strdup_printf ("%s.typelib", namespace);
 
   index = 0;
-  for (ldir = search_path; ldir; ldir = ldir->next)
+  tmp_path = build_search_path_with_overrides ();  
+  for (ldir = tmp_path; ldir; ldir = ldir->next)
     {
       GDir *dir;
       const char *dirname;
@@ -1000,18 +1063,19 @@ find_namespace_latest (const gchar  *namespace,
       candidates = g_slist_sort (candidates, (GCompareFunc) compare_candidate_reverse);
       
       elected = (struct NamespaceVersionCandidadate *) candidates->data;
-      /* Remove the elected one so we don't try to free it */
+      /* Remove the elected one so we don't try to free its contents */
       candidates = g_slist_delete_link (candidates, candidates);
       
       result = elected->mfile;
       *path_ret = elected->path;
       *version_ret = elected->version;
+      g_free (elected); /* just free the container */
       g_slist_foreach (candidates, (GFunc) free_candidate, NULL);
       g_slist_free (candidates);
     }  
-
   g_free (namespace_dash);
   g_free (namespace_typelib);
+  g_slist_free (tmp_path);  
   return result;
 }
 
@@ -1129,6 +1193,32 @@ g_irepository_require (GIRepository  *repository,
   return ret; 
 }
 
+static gboolean
+g_irepository_introspect_cb (const char *option_name,
+			     const char *value,
+			     gpointer data,
+			     GError **error)
+{
+  gboolean ret = g_irepository_dump (value, error);
+  exit (ret ? 0 : 1);
+}
+
+static const GOptionEntry introspection_args[] = {
+  { "introspect-dump", 0, G_OPTION_FLAG_HIDDEN, G_OPTION_ARG_CALLBACK,
+    g_irepository_introspect_cb, "Dump introspection information",
+    "infile.txt,outfile.xml" },
+  { NULL }
+};
+
+GOptionGroup *
+g_irepository_get_option_group (void)
+{
+  GOptionGroup *group;
+  group = g_option_group_new ("girepository", "Introspection Options", "Show Introspection Options", NULL, NULL);
+
+  g_option_group_add_entries (group, introspection_args);
+  return group;
+}
 
 GQuark
 g_irepository_error_quark (void)
