@@ -28,6 +28,12 @@
 #include "girnode.h"
 #include "gtypelib.h"
 
+struct _GIrParser
+{
+  gchar **includes;
+  GList *parsed_modules; /* All previously parsed modules */
+};
+
 typedef enum
 {
   STATE_START,    
@@ -48,7 +54,7 @@ typedef enum
   STATE_INTERFACE_PROPERTY,   /* 15 */
   STATE_INTERFACE_FIELD,
   STATE_IMPLEMENTS, 
-  STATE_REQUIRES,
+  STATE_PREREQUISITE,
   STATE_BOXED,  
   STATE_BOXED_FIELD, /* 20 */
   STATE_STRUCT,   
@@ -66,15 +72,16 @@ typedef enum
 typedef struct _ParseContext ParseContext;
 struct _ParseContext
 {
+  GIrParser *parser;
+
   ParseState state;
   ParseState prev_state;
 
-  const char * const*includes;
-  
   GList *modules;
-  gboolean prefix_aliases;
+  GList *include_modules;
   GList *dependencies;
   GHashTable *aliases;
+  GHashTable *disguised_structures;
 
   const char *namespace;
   GIrModule *current_module;
@@ -86,6 +93,34 @@ struct _ParseContext
   int type_depth;
 };
 
+static void start_element_handler (GMarkupParseContext *context,
+				   const gchar         *element_name,
+				   const gchar        **attribute_names,
+				   const gchar        **attribute_values,
+				   gpointer             user_data,
+				   GError             **error);
+static void end_element_handler   (GMarkupParseContext *context,
+				   const gchar         *element_name,
+				   gpointer             user_data,
+				   GError             **error);
+static void text_handler          (GMarkupParseContext *context,
+				   const gchar         *text,
+				   gsize                text_len,
+				   gpointer             user_data,
+				   GError             **error);
+static void cleanup               (GMarkupParseContext *context,
+				   GError              *error,
+				   gpointer             user_data);
+
+static GMarkupParser markup_parser = 
+{
+  start_element_handler,
+  end_element_handler,
+  text_handler,
+  NULL,
+  cleanup
+};
+
 static gboolean
 start_alias (GMarkupParseContext *context,
 	     const gchar         *element_name,
@@ -94,6 +129,43 @@ start_alias (GMarkupParseContext *context,
 	     ParseContext        *ctx,
 	     GError             **error);
 
+static const gchar *find_attribute (const gchar  *name, 
+				    const gchar **attribute_names,
+				    const gchar **attribute_values);
+
+
+GIrParser *
+g_ir_parser_new (void)
+{
+  GIrParser *parser = g_slice_new0 (GIrParser);
+
+  return parser;
+}
+
+void
+g_ir_parser_free (GIrParser *parser)
+{
+  GList *l;
+
+  if (parser->includes)
+    g_strfreev (parser->includes);
+
+  for (l = parser->parsed_modules; l; l = l->next)
+    g_ir_module_free (l->data);
+
+  g_slice_free (GIrParser, parser);
+}
+
+void
+g_ir_parser_set_includes (GIrParser          *parser,
+			  const gchar *const *includes)
+{
+  if (parser->includes)
+    g_strfreev (parser->includes);
+
+  parser->includes = g_strdupv ((char **)includes);
+}
+
 static void
 firstpass_start_element_handler (GMarkupParseContext *context,
 				 const gchar         *element_name,
@@ -109,6 +181,22 @@ firstpass_start_element_handler (GMarkupParseContext *context,
       start_alias (context, element_name, attribute_names, attribute_values,
 		   ctx, error);
     }
+  else if (strcmp (element_name, "record") == 0)
+    {
+      const gchar *name;
+      const gchar *disguised;
+
+      name = find_attribute ("name", attribute_names, attribute_values);
+      disguised = find_attribute ("disguised", attribute_names, attribute_values);
+
+      if (disguised && strcmp (disguised, "1") == 0)
+	{
+	  char *key;
+
+	  key = g_strdup_printf ("%s.%s", ctx->namespace, name);
+	  g_hash_table_replace (ctx->disguised_structures, key, GINT_TO_POINTER (1));
+	}
+    }
 }
 
 static void
@@ -129,7 +217,9 @@ static GMarkupParser firstpass_parser =
 };
 
 static char *
-locate_gir (const char *name, const char *version, const char * const* extra_paths)
+locate_gir (GIrParser  *parser,
+	    const char *name,
+	    const char *version)
 {
   const gchar *const *datadirs;
   const gchar *const *dir;
@@ -140,17 +230,9 @@ locate_gir (const char *name, const char *version, const char * const* extra_pat
       
   girname = g_strdup_printf ("%s-%s.gir", name, version);
   
-  for (dir = datadirs; *dir; dir++) 
+  if (parser->includes != NULL)
     {
-      path = g_build_filename (*dir, "gir", girname, NULL);
-      if (g_file_test (path, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_REGULAR))
-	return path;
-      g_free (path);
-      path = NULL;
-    }
-  if (extra_paths != NULL)
-    {
-      for (dir = extra_paths; *dir; dir++) 
+      for (dir = (const gchar *const *)parser->includes; *dir; dir++) 
 	{
 	  path = g_build_filename (*dir, girname, NULL);
 	  if (g_file_test (path, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_REGULAR))
@@ -159,6 +241,14 @@ locate_gir (const char *name, const char *version, const char * const* extra_pat
 	  path = NULL;
 	}
     }
+  for (dir = datadirs; *dir; dir++) 
+    {
+      path = g_build_filename (*dir, "gir-1.0", girname, NULL);
+      if (g_file_test (path, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_REGULAR))
+	return path;
+      g_free (path);
+      path = NULL;
+    }
   g_free (girname);
   return path;
 }
@@ -420,19 +510,64 @@ resolve_aliases (ParseContext *ctx, const gchar *type)
   gpointer orig;
   gpointer value;
   GSList *seen_values = NULL;
+  const gchar *lookup;
+  gchar *prefixed;
 
-  seen_values = g_slist_prepend (seen_values, (char*)type);
-  while (g_hash_table_lookup_extended (ctx->aliases, type, &orig, &value))
+  if (strchr (type, '.') == NULL)
     {
-      g_debug ("Resolved: %s => %s", type, (char*)value);
-      type = value;
-      if (g_slist_find_custom (seen_values, type,
+      prefixed = g_strdup_printf ("%s.%s", ctx->namespace, type);
+      lookup = prefixed;
+    }
+  else
+    {
+      lookup = type;
+      prefixed = NULL;
+    }
+
+  seen_values = g_slist_prepend (seen_values, (char*)lookup);
+  while (g_hash_table_lookup_extended (ctx->current_module->aliases, lookup, &orig, &value))
+    {
+      g_debug ("Resolved: %s => %s\n", lookup, (char*)value);
+      lookup = value;
+      if (g_slist_find_custom (seen_values, lookup,
 			       (GCompareFunc)strcmp) != NULL)
 	break;
-      seen_values = g_slist_prepend (seen_values, (gchar*)type);
+      seen_values = g_slist_prepend (seen_values, (gchar*)lookup);
     }
   g_slist_free (seen_values);
-  return type;
+
+  if (lookup == prefixed)
+    lookup = type;
+
+  g_free (prefixed);
+  
+  return lookup;
+}
+
+static gboolean
+is_disguised_structure (ParseContext *ctx, const gchar *type)
+{
+  const gchar *lookup;
+  gchar *prefixed;
+  gboolean result;
+
+  if (strchr (type, '.') == NULL)
+    {
+      prefixed = g_strdup_printf ("%s.%s", ctx->namespace, type);
+      lookup = prefixed;
+    }
+  else
+    {
+      lookup = type;
+      prefixed = NULL;
+    }
+
+  result = g_hash_table_lookup (ctx->current_module->disguised_structures,
+				lookup) != NULL;
+  
+  g_free (prefixed);
+  
+  return result;
 }
 
 static GIrNodeType *
@@ -539,10 +674,12 @@ start_function (GMarkupParseContext *context,
 	       strcmp (element_name, "callback") == 0);
       break;
     case STATE_CLASS:
+      found = strcmp (element_name, "function") == 0;
+	/* fallthrough */
     case STATE_BOXED:
     case STATE_STRUCT:
     case STATE_UNION:
-      found = strcmp (element_name, "constructor") == 0;
+      found = (found || strcmp (element_name, "constructor") == 0);
       /* fallthrough */
     case STATE_INTERFACE:
       found = (found ||
@@ -700,6 +837,9 @@ start_parameter (GMarkupParseContext *context,
   const gchar *optional;
   const gchar *allow_none;
   const gchar *transfer;
+  const gchar *scope;
+  const gchar *closure;
+  const gchar *destroy;
   GIrNodeParam *param;
       
   if (!(strcmp (element_name, "parameter") == 0 &&
@@ -713,7 +853,10 @@ start_parameter (GMarkupParseContext *context,
   optional = find_attribute ("optional", attribute_names, attribute_values);
   allow_none = find_attribute ("allow-none", attribute_names, attribute_values);
   transfer = find_attribute ("transfer-ownership", attribute_names, attribute_values);
-
+  scope = find_attribute ("scope", attribute_names, attribute_values);
+  closure = find_attribute ("closure", attribute_names, attribute_values);
+  destroy = find_attribute ("destroy", attribute_names, attribute_values);
+  
   if (name == NULL)
     name = "unknown";
 
@@ -762,6 +905,20 @@ start_parameter (GMarkupParseContext *context,
 
   parse_param_transfer (param, transfer);
 
+  if (scope && strcmp (scope, "call") == 0)
+    param->scope = GI_SCOPE_TYPE_CALL;
+  else if (scope && strcmp (scope, "object") == 0)
+    param->scope = GI_SCOPE_TYPE_OBJECT;
+  else if (scope && strcmp (scope, "async") == 0)
+    param->scope = GI_SCOPE_TYPE_ASYNC;
+  else if (scope && strcmp (scope, "notified") == 0)
+    param->scope = GI_SCOPE_TYPE_NOTIFIED;
+  else
+    param->scope = GI_SCOPE_TYPE_INVALID;
+  
+  param->closure = closure ? atoi (closure) : -1;
+  param->destroy = destroy ? atoi (destroy) : -1;
+  
   ((GIrNode *)param)->name = g_strdup (name);
 	  
   switch (ctx->current_node->type)
@@ -811,7 +968,6 @@ start_field (GMarkupParseContext *context,
   const gchar *writable;
   const gchar *bits;
   const gchar *branch;
-  const gchar *offset;
   GIrNodeField *field;
 
   switch (ctx->state)
@@ -834,7 +990,6 @@ start_field (GMarkupParseContext *context,
   writable = find_attribute ("writable", attribute_names, attribute_values);
   bits = find_attribute ("bits", attribute_names, attribute_values);
   branch = find_attribute ("branch", attribute_names, attribute_values);
-  offset = find_attribute ("offset", attribute_names, attribute_values);
   
   if (name == NULL)
     {
@@ -856,11 +1011,6 @@ start_field (GMarkupParseContext *context,
   else
     field->bits = 0;
   
-  if (offset)
-    field->offset = atoi (offset);
-  else
-    field->offset = 0;
-  
   switch (ctx->current_node->type)
     {
     case G_IR_NODE_OBJECT:
@@ -955,25 +1105,18 @@ start_alias (GMarkupParseContext *context,
     }
 
   value = g_strdup (target);
-  if (ctx->prefix_aliases)
+  key = g_strdup_printf ("%s.%s", ctx->namespace, name);
+  if (!strchr (target, '.'))
     {
-      key = g_strdup_printf ("%s.%s", ctx->namespace, name);
-      if (!strchr (target, '.'))
+      const BasicTypeInfo *basic = parse_basic (target);
+      if (!basic)
 	{
-	  const BasicTypeInfo *basic = parse_basic (target);
-	  if (!basic)
-	    {
-	      g_free (value);
-	      /* For non-basic types, re-qualify the interface */
-	      value = g_strdup_printf ("%s.%s", ctx->namespace, target);
-	    }
+	  g_free (value);
+	  /* For non-basic types, re-qualify the interface */
+	  value = g_strdup_printf ("%s.%s", ctx->namespace, target);
 	}
     }
-  else
-    {
-      key = g_strdup (name);
-    }
-  g_hash_table_insert (ctx->aliases, key, value);
+  g_hash_table_replace (ctx->aliases, key, value);
 
   return TRUE;
 }
@@ -1535,22 +1678,37 @@ start_type (GMarkupParseContext *context,
     }
   else
     {
-      gboolean is_pointer;
+      int pointer_depth;
       name = find_attribute ("name", attribute_names, attribute_values);
 
       if (name == NULL)
 	MISSING_ATTRIBUTE (context, error, element_name, "name");
       
+      pointer_depth = 0;
       ctype = find_attribute ("c:type", attribute_names, attribute_values);
-      if (ctype != NULL && strchr (ctype, '*'))
-	is_pointer = TRUE;
-      else
-	is_pointer = FALSE;
-
+      if (ctype != NULL)
+        {
+          const char *cp = ctype + strlen(ctype) - 1;
+          while (cp > ctype && *cp-- == '*')
+            pointer_depth++;
+        }
+      
+      if (ctx->current_typed->type == G_IR_NODE_PARAM &&
+          ((GIrNodeParam *)ctx->current_typed)->out &&
+          pointer_depth > 0)
+        pointer_depth--;
+      
       typenode = parse_type (ctx, name);
 
-      if (is_pointer)
-	typenode->is_pointer = is_pointer;
+      /* A 'disguised' structure is one where the c:type is a typedef that
+       * doesn't look like a pointer, but is internally.
+       */
+      if (typenode->tag == GI_TYPE_TAG_INTERFACE &&
+	  is_disguised_structure (ctx, typenode->interface))
+	pointer_depth++;
+
+      if (pointer_depth > 0)
+	typenode->is_pointer = TRUE;
     }
 
   ctx->type_parameters = g_list_append (ctx->type_parameters, typenode);
@@ -1936,12 +2094,14 @@ start_struct (GMarkupParseContext *context,
     {
       const gchar *name;
       const gchar *deprecated;
+      const gchar *disguised;
       const gchar *gtype_name;
       const gchar *gtype_init;
       GIrNodeStruct *struct_;
       
       name = find_attribute ("name", attribute_names, attribute_values);
       deprecated = find_attribute ("deprecated", attribute_names, attribute_values);
+      disguised = find_attribute ("disguised", attribute_names, attribute_values);
       gtype_name = find_attribute ("glib:type-name", attribute_names, attribute_values);
       gtype_init = find_attribute ("glib:get-type", attribute_names, attribute_values);
 
@@ -1969,6 +2129,9 @@ start_struct (GMarkupParseContext *context,
       else
 	struct_->deprecated = FALSE;
 
+      if (disguised && strcmp (disguised, "1") == 0)
+	struct_->disguised = TRUE;
+
       struct_->gtype_name = g_strdup (gtype_name);
       struct_->gtype_init = g_strdup (gtype_init);
       
@@ -2071,12 +2234,38 @@ parse_include (GMarkupParseContext *context,
 	       const char          *version,
 	       GError             **error)
 {
-  ParseContext sub_ctx = { 0 };
   gchar *buffer;
   gsize length;
   char *girpath;
-  
-  girpath = locate_gir (name, version, ctx->includes);
+  gboolean success = FALSE;
+  GList *modules;
+  GList *l;
+
+  for (l = ctx->parser->parsed_modules; l; l = l->next)
+    {
+      GIrModule *m = l->data;
+
+      if (strcmp (m->name, name) == 0)
+	{
+	  if (strcmp (m->version, version) == 0)
+	    {
+	      ctx->include_modules = g_list_prepend (ctx->include_modules, m);
+
+	      return TRUE;
+	    }
+	  else
+	    {
+	      g_set_error (error,
+			   G_MARKUP_ERROR,
+			   G_MARKUP_ERROR_INVALID_CONTENT,
+			   "Module '%s' imported with conflicting versions '%s' and '%s'",
+			   name, m->version, version);
+	      return FALSE;
+	    }
+	}
+    }
+
+  girpath = locate_gir (ctx->parser, name, version);
 
   if (girpath == NULL)
     {
@@ -2097,29 +2286,15 @@ parse_include (GMarkupParseContext *context,
     }
   g_free (girpath);
 
-  sub_ctx.state = STATE_START;
-  sub_ctx.includes = ctx->includes;
-  sub_ctx.prefix_aliases = TRUE;
-  sub_ctx.namespace = name;
-  sub_ctx.aliases = ctx->aliases;
-  sub_ctx.type_depth = 0;
+  modules = g_ir_parser_parse_string (ctx->parser, name, buffer, length, error);
+  success = error != NULL;
 
-  context = g_markup_parse_context_new (&firstpass_parser, 0, &sub_ctx, NULL);
-	  
-  if (!g_markup_parse_context_parse (context, buffer, length, error))
-    {
-      g_free (buffer);
-      return FALSE;
-    }
-	  
-  if (!g_markup_parse_context_end_parse (context, error))
-    {
-      g_free (buffer);
-      return FALSE;
-    }
-	  
-  g_markup_parse_context_free (context);
-  return TRUE;
+  ctx->include_modules = g_list_concat (ctx->include_modules,
+					modules);
+
+  g_free (buffer);
+
+  return success;
 }
   
 extern GLogLevelFlags logged_levels;
@@ -2185,25 +2360,6 @@ start_element_handler (GMarkupParseContext *context,
 			    attribute_names, attribute_values,
 			    ctx, error))
 	goto out;
-      else if (strcmp (element_name, "class") == 0 &&
-	       ctx->state == STATE_REQUIRES)
-	{
-	  const gchar *name;
-
-	  name = find_attribute ("name", attribute_names, attribute_values);
-
-	  if (name == NULL)
-	    MISSING_ATTRIBUTE (context, error, element_name, "name");
-	  else
-	    {  
-	      GIrNodeInterface *iface;
-
-	      iface = (GIrNodeInterface *)ctx->current_node;
-	      iface ->prerequisites = g_list_append (iface->prerequisites, g_strdup (name));
-	    }
-
-	  goto out;
-	}
       break;
 
     case 'd':
@@ -2302,7 +2458,16 @@ start_element_handler (GMarkupParseContext *context,
       if (strcmp (element_name, "namespace") == 0 && ctx->state == STATE_REPOSITORY)
 	{
 	  const gchar *name, *version, *shared_library;
-	  
+
+	  if (ctx->current_module != NULL)
+	    {
+	      g_set_error (error,
+			   G_MARKUP_ERROR,
+			   G_MARKUP_ERROR_INVALID_CONTENT,
+			   "Only one <namespace/> element is currently allowed per <repository/>");
+	      goto out;
+	    }
+
 	  name = find_attribute ("name", attribute_names, attribute_values);
 	  version = find_attribute ("version", attribute_names, attribute_values);
 	  shared_library = find_attribute ("shared-library", attribute_names, attribute_values);
@@ -2313,7 +2478,28 @@ start_element_handler (GMarkupParseContext *context,
 	    MISSING_ATTRIBUTE (context, error, element_name, "version");
 	  else
 	    {
+	      GList *l;
+
+	      if (strcmp (name, ctx->namespace) != 0)
+		g_set_error (error,
+			     G_MARKUP_ERROR,
+			     G_MARKUP_ERROR_INVALID_CONTENT,
+			     "<namespace/> name element '%s' doesn't match file name '%s'",
+			     name, ctx->namespace);
+
 	      ctx->current_module = g_ir_module_new (name, version, shared_library);
+
+	      ctx->current_module->aliases = ctx->aliases;
+	      ctx->aliases = NULL;
+	      ctx->current_module->disguised_structures = ctx->disguised_structures;
+	      ctx->disguised_structures = NULL;
+
+	      for (l = ctx->include_modules; l; l = l->next)
+		g_ir_module_add_include_module (ctx->current_module, l->data);
+
+	      g_list_free (ctx->include_modules);
+	      ctx->include_modules = NULL;
+
 	      ctx->modules = g_list_append (ctx->modules, ctx->current_module);
 	      ctx->current_module->dependencies = ctx->dependencies;
 
@@ -2339,7 +2525,26 @@ start_element_handler (GMarkupParseContext *context,
 				attribute_names, attribute_values,
 				ctx, error))
 	goto out;
+      else if (strcmp (element_name, "prerequisite") == 0 &&
+	       ctx->state == STATE_INTERFACE)
+	{
+	  const gchar *name;
 
+	  name = find_attribute ("name", attribute_names, attribute_values);
+
+	  state_switch (ctx, STATE_PREREQUISITE);
+
+	  if (name == NULL)
+	    MISSING_ATTRIBUTE (context, error, element_name, "name");
+	  else
+	    {  
+	      GIrNodeInterface *iface;
+
+	      iface = (GIrNodeInterface *)ctx->current_node;
+	      iface->prerequisites = g_list_append (iface->prerequisites, g_strdup (name));
+	    }
+	  goto out;
+	}
       break;
 
     case 'r':
@@ -2366,13 +2571,6 @@ start_element_handler (GMarkupParseContext *context,
 				   attribute_names, attribute_values,
 				   ctx, error))
 	goto out;      
-      else if (strcmp (element_name, "requires") == 0 &&
-	       ctx->state == STATE_INTERFACE)
-	{
-	  state_switch (ctx, STATE_REQUIRES);
-	  
-	  goto out;
-	}
       else if (start_struct (context, element_name,
 			     attribute_names, attribute_values,
 			     ctx, error))
@@ -2710,8 +2908,8 @@ end_element_handler (GMarkupParseContext *context,
       if (require_end_element (context, ctx, "implements", element_name, error))
         state_switch (ctx, STATE_CLASS);
       break;
-    case STATE_REQUIRES:
-      if (require_end_element (context, ctx, "requires", element_name, error))
+    case STATE_PREREQUISITE:
+      if (require_end_element (context, ctx, "prerequisite", element_name, error))
         state_switch (ctx, STATE_INTERFACE);
       break;
     case STATE_NAMESPACE_CONSTANT:
@@ -2777,15 +2975,6 @@ cleanup (GMarkupParseContext *context,
   ctx->current_module = NULL;
 }
 
-static GMarkupParser parser = 
-{
-  start_element_handler,
-  end_element_handler,
-  text_handler,
-  NULL,
-  cleanup
-};
-
 static GList *
 post_filter_varargs_functions (GList *list)
 {
@@ -2847,21 +3036,33 @@ post_filter (GIrModule *module)
     }
 }
 
-GList * 
-g_ir_parse_string (const gchar  *namespace,
-		   const gchar *const *includes,
-		   const gchar  *buffer, 
-		   gssize        length,
-		   GError      **error)
+/**
+ * g_ir_parser_parse_string:
+ * @parser: a #GIrParser
+ * @error: return location for a #GError, or %NULL
+ *
+ * Parse a string that holds a complete GIR XML file, and return a list of a
+ * a #GirModule for each &lt;namespace/&gt; element within the file.
+ *
+ * @returns: a newly allocated list of #GIrModule. The modules themselves
+ *  are owned by the #GIrParser and will be freed along with the parser.
+ */
+GList *
+g_ir_parser_parse_string (GIrParser           *parser,
+			  const gchar         *namespace,
+			  const gchar         *buffer,
+			  gssize               length,
+			  GError             **error)
 {
   ParseContext ctx = { 0 };
   GMarkupParseContext *context;
 
+  ctx.parser = parser;
   ctx.state = STATE_START;
-  ctx.includes = includes;
-  ctx.prefix_aliases = FALSE;
   ctx.namespace = namespace;
+  ctx.include_modules = NULL;
   ctx.aliases = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
+  ctx.disguised_structures = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
   ctx.type_depth = 0;
   ctx.dependencies = NULL;
   ctx.current_module = NULL;
@@ -2876,32 +3077,57 @@ g_ir_parse_string (const gchar  *namespace,
 
   g_markup_parse_context_free (context);
   
-  context = g_markup_parse_context_new (&parser, 0, &ctx, NULL);
+  context = g_markup_parse_context_new (&markup_parser, 0, &ctx, NULL);
   if (!g_markup_parse_context_parse (context, buffer, length, error))
     goto out;
 
   if (!g_markup_parse_context_end_parse (context, error))
     goto out;
 
+  parser->parsed_modules = g_list_concat (g_list_copy (ctx.modules),
+					  parser->parsed_modules);
+
  out:
 
-  g_hash_table_destroy (ctx.aliases);
+  if (ctx.modules == NULL)
+    {
+      /* An error occurred before we created a module, so we haven't
+       * transferred ownership of these hash tables to the module.
+       */
+      if (ctx.aliases != NULL)
+	g_hash_table_destroy (ctx.aliases);
+      if (ctx.disguised_structures != NULL)
+	g_hash_table_destroy (ctx.disguised_structures);
+      g_list_free (ctx.include_modules);
+    }
   
   g_markup_parse_context_free (context);
   
   return ctx.modules;
 }
 
+/**
+ * g_ir_parser_parse_file:
+ * @parser: a #GIrParser
+ * @error: return location for a #GError, or %NULL
+ *
+ * Parse GIR XML file, and return a list of a a #GirModule for each
+ * &lt;namespace/&gt; element within the file.
+ *
+ * @returns: a newly allocated list of #GIrModule. The modules themselves
+ *  are owned by the #GIrParser and will be freed along with the parser.
+ */
 GList *
-g_ir_parse_file (const gchar  *filename,
-		 const gchar *const *includes,
-		 GError      **error)
+g_ir_parser_parse_file (GIrParser   *parser,
+			const gchar *filename,
+			GError     **error)
 {
   gchar *buffer;
   gsize length;
   GList *modules;
   GList *iter;
   const char *slash;
+  char *dash;
   char *namespace;
 
   if (!g_str_has_suffix (filename, ".gir"))
@@ -2922,10 +3148,15 @@ g_ir_parse_file (const gchar  *filename,
     namespace = g_strdup (slash+1);
   namespace[strlen(namespace)-4] = '\0';
 
+  /* Remove version */
+  dash = strstr (namespace, "-");
+  if (dash != NULL)
+    *dash = '\0';
+
   if (!g_file_get_contents (filename, &buffer, &length, error))
     return NULL;
   
-  modules = g_ir_parse_string (namespace, includes, buffer, length, error);
+  modules = g_ir_parser_parse_string (parser, namespace, buffer, length, error);
 
   for (iter = modules; iter; iter = iter->next) 
     {
